from __future__ import annotations

from typing import Any

from django.conf import settings
from django.db.models import Field, Model

from .utils import get_s3_path_from_url
from .widgets import S3UploadWidget


class S3UploadField(Field):
    def __init__(self, dest: str | None = None, *args: Any, **kwargs: Any) -> None:
        if not dest:
            raise ValueError("S3UploadField must be initialised with a destination")
        self.dest = dest
        self.widget = S3UploadWidget(self.dest)
        super(S3UploadField, self).__init__(*args, **kwargs)

    def deconstruct(self) -> tuple[str, str, Any, Any]:
        name, path, args, kwargs = super(S3UploadField, self).deconstruct()
        kwargs["dest"] = self.dest
        return name, path, args, kwargs

    def get_internal_type(self) -> str:
        return "TextField"

    def formfield(self, *args: Any, **kwargs: Any) -> Any:
        kwargs["widget"] = self.widget
        return super(S3UploadField, self).formfield(*args, **kwargs)

    def pre_save(self, model_instance: Model, add: bool) -> str:
        file_url = getattr(model_instance, self.attname)

        if file_url:
            setattr(model_instance, self.attname, file_url)
            bucket_name = settings.S3UPLOAD_DESTINATIONS[self.dest].get(
                "bucket", settings.AWS_STORAGE_BUCKET_NAME
            )
            return get_s3_path_from_url(file_url, bucket_name=bucket_name)

        return file_url



from __future__ import annotations

import hashlib
import hmac
import json
from base64 import b64encode
from datetime import datetime, timedelta, timezone
from typing import Any
from urllib.parse import unquote, urlparse

import boto3
from botocore.config import Config
from django.conf import settings


def create_upload_data(  # noqa: C901
    *,
    content_type: str,
    key: str,
    acl: str,
    bucket: str | None = None,
    cache_control: str | None = None,
    content_disposition: str | None = None,
    content_length_range: str | None = None,
    server_side_encryption: str | None = None,
    token: str | None = None,
) -> dict[str, Any]:
    """Generate AWS upload payload."""
    access_key = settings.AWS_ACCESS_KEY_ID
    secret_access_key = settings.AWS_SECRET_ACCESS_KEY
    bucket = bucket or settings.AWS_STORAGE_BUCKET_NAME
    region = settings.S3UPLOAD_REGION
    bucket_url = get_bucket_endpoint_url(bucket, region)
    expires_in = datetime.now(timezone.utc) + timedelta(seconds=60 * 5)
    expires = expires_in.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    now_date = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%S000Z")
    raw_date = datetime.now(timezone.utc).strftime("%Y%m%d")

    policy_dict: dict[str, Any] = {
        "expiration": expires,
        "conditions": [
            {"bucket": bucket},
            {"acl": acl},
            ["starts-with", "$key", ""],
            {"success_action_status": "201"},
            {"x-amz-credential": f"{access_key}/{raw_date}/{region}/s3/aws4_request"},
            {"x-amz-algorithm": "AWS4-HMAC-SHA256"},
            {"x-amz-date": now_date},
            {"content-type": content_type},
        ],
    }

    if token:
        policy_dict["conditions"].append({"x-amz-security-token": token})

    if cache_control:
        policy_dict["conditions"].append({"Cache-Control": cache_control})

    if content_disposition:
        policy_dict["conditions"].append({"Content-Disposition": content_disposition})

    if server_side_encryption:
        policy_dict["conditions"].append(
            {"x-amz-server-side-encryption": server_side_encryption}
        )

    if content_length_range:
        policy_dict["conditions"].append(
            ["content-length-range", content_length_range[0], content_length_range[1]]
        )

    policy_object = json.dumps(policy_dict)

    policy = b64encode(policy_object.replace("\n", "").replace("\r", "").encode())

    date_key = hmac.new(
        b"AWS4" + secret_access_key.encode("utf-8"),
        msg=raw_date.encode("utf-8"),
        digestmod=hashlib.sha256,
    ).digest()

    date_region_key = hmac.new(
        date_key, msg=region.encode("utf-8"), digestmod=hashlib.sha256
    ).digest()

    date_region_service_key = hmac.new(
        date_region_key, msg=b"s3", digestmod=hashlib.sha256
    ).digest()

    signing_key = hmac.new(
        date_region_service_key, msg=b"aws4_request", digestmod=hashlib.sha256
    ).digest()

    signature = hmac.new(signing_key, msg=policy, digestmod=hashlib.sha256).hexdigest()
    return_dict = {
        "policy": policy.decode(),  # decode to make it JSON serializable
        "success_action_status": 201,
        "x-amz-credential": f"{access_key}/{raw_date}/{region}/s3/aws4_request",
        "x-amz-date": now_date,
        "x-amz-signature": signature,
        "x-amz-algorithm": "AWS4-HMAC-SHA256",
        "form_action": bucket_url,
        "key": key,
        "acl": acl,
        "content-type": content_type,
    }

    if token:
        return_dict["x-amz-security-token"] = token

    if server_side_encryption:
        return_dict["x-amz-server-side-encryption"] = server_side_encryption

    if cache_control:
        return_dict["Cache-Control"] = cache_control

    if content_disposition:
        return_dict["Content-Disposition"] = content_disposition

    return return_dict


def get_s3_path_from_url(
    url: str, bucket_name: str = settings.AWS_STORAGE_BUCKET_NAME
) -> str:
    decoded = unquote(url)
    path = urlparse(decoded).path

    # The bucket name might be part of the path,
    # so get the path that comes after the bucket name
    key_path = path.split(bucket_name)[-1]

    # Remove slash prefix if present
    if key_path[0] == "/":
        key_path = key_path[1:]

    return key_path


def get_signed_download_url(
    key: str,
    bucket_name: str | None = None,
    ttl: int = 60,
) -> str:
    bucket_name = bucket_name or settings.AWS_STORAGE_BUCKET_NAME
    s3 = boto3.client(
        "s3",
        aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
        aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
        config=Config(signature_version="s3v4"),
        region_name=settings.S3UPLOAD_REGION,
    )
    download_url = s3.generate_presigned_url(
        "get_object", Params={"Bucket": bucket_name, "Key": key}, ExpiresIn=ttl
    )
    return download_url


# virtual-hosted-style URLs are now the default
# Example: https://bucket-name.s3.Region.amazonaws.com/key name
# See: https://docs.aws.amazon.com/AmazonS3/latest/userguide/access-bucket-intro.html
def get_bucket_endpoint_url(
    bucket_name: str | None = None,
    region: str | None = None,
    default: str = "https://{bucket}.s3.{region}.amazonaws.com",
) -> str:
    bucket_name = bucket_name or settings.AWS_STORAGE_BUCKET_NAME
    region = region or settings.S3UPLOAD_REGION
    bucket_endpoint = getattr(
        settings,
        "S3UPLOAD_BUCKET_ENDPOINT",
        default,
    )
    return bucket_endpoint.format(bucket=bucket_name, region=region)



from __future__ import annotations

import os
from typing import Any
from urllib.parse import unquote_plus as urlunquote_plus

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.forms import widgets
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe

from .utils import (
    get_bucket_endpoint_url,
    get_s3_path_from_url,
    get_signed_download_url,
)


class S3UploadWidget(widgets.TextInput):
    class Media:
        js = ("s3upload/js/django-s3-uploads.min.js",)
        css = {
            "all": (
                "s3upload/css/bootstrap-progress.min.css",
                "s3upload/css/styles.css",
            )
        }

    def __init__(self, dest: str, **kwargs: Any) -> None:
        if not dest:
            raise ValueError("S3UploadWidget must be initialised with a destination")
        if dest not in settings.S3UPLOAD_DESTINATIONS:
            raise ImproperlyConfigured(
                "S3UploadWidget destination '%s' is not configured. "
                "Please check settings.S3UPLOAD_DESTINATIONS." % dest
            )
        self.acl = settings.S3UPLOAD_DESTINATIONS[dest].get("acl", "public-read")
        self.dest = dest
        super(S3UploadWidget, self).__init__(**kwargs)

    def get_file_url(self, value: str) -> str:
        if value:
            bucket_name = settings.S3UPLOAD_DESTINATIONS[self.dest].get("bucket")
            if self.acl == "private":
                return get_signed_download_url(value, bucket_name)
            else:
                # Default to virtual-hosted–style URL
                bucket_url = get_bucket_endpoint_url(bucket_name)
                return "{}/{}".format(bucket_url, value)
        else:
            return ""

    def get_attr(
        self, attrs: dict[str, Any] | None, key: str, default: str = ""
    ) -> str:
        return self.build_attrs(attrs).get(key, default) if attrs else default

    def render(
        self, name: str, value: str, attrs: dict[str, Any] | None = None, **kwargs: Any
    ) -> str:
        path = get_s3_path_from_url(value) if value else ""
        file_name = os.path.basename(urlunquote_plus(path))
        tpl = os.path.join("s3upload", "s3upload-widget.tpl")
        output = render_to_string(
            tpl,
            {
                "policy_url": reverse("s3upload"),
                "element_id": self.get_attr(attrs, "id"),
                "file_name": file_name,
                "dest": self.dest,
                "file_url": self.get_file_url(path),
                "name": name,
                "style": self.get_attr(attrs, "style"),
            },
        )
        # TODO: review use of mark_safe - will the template render
        # cover cases of bad filenames?
        return mark_safe(output)  # noqa: S308



from __future__ import annotations

from os.path import splitext

from django.conf import settings
from django.http import HttpRequest, JsonResponse
from django.utils.text import get_valid_filename
from django.views.decorators.http import require_POST

from .utils import create_upload_data, get_signed_download_url


@require_POST
def get_upload_params(request: HttpRequest) -> JsonResponse:  # noqa: C901
    content_type = request.POST["type"]
    filename = get_valid_filename(request.POST["name"])
    dest = settings.S3UPLOAD_DESTINATIONS[request.POST["dest"]]

    if not dest:
        return JsonResponse({"error": "File destination does not exist."}, status=400)

    key = dest.get("key")
    auth = dest.get("auth")
    allowed_types = dest.get("allowed_types")
    acl = dest.get("acl")
    bucket = dest.get("bucket")
    cache_control = dest.get("cache_control")
    content_disposition = dest.get("content_disposition")
    content_length_range = dest.get("content_length_range")
    allowed_extensions = dest.get("allowed_extensions")
    server_side_encryption = dest.get("server_side_encryption")

    if not acl:
        acl = "public-read"

    if not key:
        return JsonResponse({"error": "Missing destination path."}, status=403)

    if auth and not auth(request.user):
        return JsonResponse({"error": "Permission denied."}, status=403)

    if (allowed_types and content_type not in allowed_types) and allowed_types != "*":
        return JsonResponse(
            {"error": "Invalid file type (%s)." % content_type}, status=400
        )

    original_ext = splitext(filename)[1]
    lowercased_ext = original_ext.lower()
    if (
        allowed_extensions and lowercased_ext not in allowed_extensions
    ) and allowed_extensions != "*":
        return JsonResponse(
            {"error": "Forbidden file extension (%s)." % original_ext}, status=415
        )

    if hasattr(key, "__call__"):
        key = key(filename)
    elif key == "/":
        key = filename
    else:
        key = "{}/{}".format(key, filename)

    aws_payload = create_upload_data(
        content_type=content_type,
        key=key,
        acl=acl,
        bucket=bucket,
        cache_control=cache_control,
        content_disposition=content_disposition,
        content_length_range=content_length_range,
        server_side_encryption=server_side_encryption,
    )

    url = None

    # Generate signed URL for private document access
    if acl == "private":
        url = get_signed_download_url(
            key=key.replace("${filename}", filename),
            bucket_name=bucket,
            ttl=int(5 * 60),  # 5 mins
        )

    data = {
        "aws_payload": aws_payload,
        "private_access_url": url,
    }

    return JsonResponse(data, content_type="application/json")
